import {
  Inject,
  Injectable,
  ForbiddenException,
  UnauthorizedException,
} from '@nestjs/common'
import * as bcrypt from 'bcrypt'
import { PrismaService } from 'nestjs-prisma'
import { JwtService } from '@nestjs/jwt'
import { securityConfig, SecurityConfig } from 'src/common/configs'
import { EmailScene } from 'src/email/dto/email.dto'
import { DeleteUserDto } from './dto/delete-user.dto'
import { ChangePassword } from './dto/change-password.dto'
import { TokenContnet } from './dto/token.dto'
import { ChangeEmailDto } from './dto/change-email.dto'
import { VerifyEmail } from 'src/email/email-verify.decorator'

@Injectable()
export class MeService {
  constructor(
    private jwtService: JwtService,
    private prismaService: PrismaService,
    @Inject(securityConfig.KEY) private secureConfig: SecurityConfig,
  ) {}

  @VerifyEmail(EmailScene.deleteUser)
  async deleteUser(userToDelete: DeleteUserDto, userId: string) {
    const { password } = userToDelete
    const user = await this.prismaService.user.findUniqueOrThrow({
      where: { email: userToDelete.email },
    })
    if (user.id !== userId) {
      throw new ForbiddenException()
    }
    const passwordValid = await bcrypt.compare(password, user.password)
    if (!passwordValid) {
      throw new ForbiddenException('Invalid password')
    }

    return this.prismaService.user.delete({
      where: { email: userToDelete.email },
    })
  }

  async changePassword(payload: ChangePassword, userId: string) {
    const user = await this.prismaService.user.findUniqueOrThrow({
      where: { id: userId },
    })
    await this.checkPassword(payload.oldPassword, user.password)
    const hashedPassword = await bcrypt.hash(
      payload.newPassword,
      this.secureConfig.bcryptSaltOrRound,
    )
    return this.prismaService.user.update({
      where: { id: userId },
      data: { password: hashedPassword },
    })
  }

  @VerifyEmail(EmailScene.changeEmail)
  async changeEmail(payload: ChangeEmailDto, userId: string) {
    const { email } = payload
    return this.prismaService.user.update({
      where: { id: userId },
      data: { email },
    })
  }

  async updateAccessToken(refreshToken: string | undefined) {
    if (!refreshToken) {
      throw new UnauthorizedException('没有令牌，请重新登录')
    }
    const { userId, iat } = this.jwtService.verify<TokenContnet>(refreshToken, {
      secret: this.secureConfig.jwt_refresh_secret,
    })
    const user = await this.prismaService.user.findUniqueOrThrow({
      where: { id: userId },
    })
    // TODO：不使用updatedAt，而是自定义的一个refreshTime字段
    if (iat * 1000 < user.updatedAt.getTime()) {
      throw new UnauthorizedException('令牌失效，请重新登录')
    }
    const accessToken = this.jwtService.sign(
      { userId },
      {
        secret: this.secureConfig.jwt_access_secret,
        expiresIn: this.secureConfig.expiresIn,
      },
    )
    return accessToken
  }

  private async checkPassword(pwd: string, hashPwd: string) {
    const valid = await bcrypt.compare(pwd, hashPwd)
    if (!valid) {
      throw new ForbiddenException('Invalid password')
    }
  }
}
